#include "base64tx.h"

uint8_t base64_buffer[32];
uint8_t bits_in_buffer;

void push_value(uint32_t value, uint8_t bits)
{
    uint8_t remaining_bits = bits;
    uint8_t free_bits_in_current_element = 8 - (bits_in_buffer % 8);
    uint16_t current_buffer_element = bits_in_buffer / 8;
    bits_in_buffer += bits;
    
    // Push the highest bits of value in the first non-full buffer element.
    // That are 1..8 bits:
    if (bits <= free_bits_in_current_element) {
        // This value fits completely into the remaining space of the last element.
        // Consequently, value is only one byte.
        // Shift it left, so it's highest bit touches the last bit in the buffer:
        base64_buffer[current_buffer_element] |=
            (uint8_t)value << (free_bits_in_current_element - bits);
        return;
    }
    
    // If this value doesn'f fit completely, shift it right, so its
    // highest bit touches the last value:
    uint8_t shift = bits - free_bits_in_current_element;
    uint32_t value_shifted = value >> (shift % 8);
    uint8_t high_bits;
    if (shift < 8) {
        high_bits = value_shifted;
    } else if (shift < 16) {
        high_bits = value_shifted >> 8;
    } else if (shift < 24) {
        high_bits = value_shifted >> 16;
    } else {
        high_bits = value_shifted >> 24;
    }
    base64_buffer[current_buffer_element++] |= high_bits;
    remaining_bits -= free_bits_in_current_element;
    
    // As we are at a byte boundary now, shift value left, so its highest
    // remaining bit is at a byte boundary.
    // The lowest bit is still in the lowest byte
    value <<= 8 - (remaining_bits % 8);
    uint8_t highest_byte_in_value = remaining_bits / 8;
    
    // Copy the remaining bytes of value to the buffer:
    // (remaining_bits says how many bytes are left in value)
    switch (highest_byte_in_value) {
        case 3 : base64_buffer[current_buffer_element++] = value >> 24;
        case 2 : base64_buffer[current_buffer_element++] = value >> 16;
        case 1 : base64_buffer[current_buffer_element++] = value >> 8;
        default : base64_buffer[current_buffer_element] = value;
    }
}

void send_buffer_base64_no_newline(void)
{
    // Read from buffer in chunks of 3 bytes, and transmit them as 4 Base64 characters.
    // When bits_in_buffer is reached, only the needed characters are transmitted,
    // the end is padded with zero bits.
    
    uint8_t chunk_low_byte = 0;
    uint8_t remaining_bits = bits_in_buffer;
    
    while (1) {
        // Read the next chunk of bytes:
        uint8_t byte0 = base64_buffer[chunk_low_byte];
        uint8_t byte1 = base64_buffer[chunk_low_byte + 1];
        uint8_t byte2 = base64_buffer[chunk_low_byte + 2];

        // Send the bytes as Base64.
        // The last byte with data is padded with zero.
        // So, if data ends in the byte, send it without the next byte and break.
        uint8_t char0 = byte0 >> 2;
        send_base64_char(char0);
        
        uint8_t char1 = byte0 << 4;
        if (remaining_bits <= 8) {
            send_base64_char(char1);
            break;
        }
        char1 |= byte1 >> 4;
        send_base64_char(char1);
        
        uint8_t char2 = byte1 << 2;
        if (remaining_bits <= 16) {
            send_base64_char(char2);
            break;
        }
        char2 |= byte2 >> 6;
        send_base64_char(char2);
        
        uint8_t char3 = byte2;
        if (remaining_bits <= 18) {
            break;
        }
        send_base64_char(char3);
        
        // Done with this chunk, test whether there is more data:
        if (remaining_bits <= 24) {
            break;
        } else {
            chunk_low_byte += 3;
            remaining_bits -= 24;
        }
    }
}

void reset_buffer(void)
{
    bits_in_buffer = 0;
    for (uint16_t i = 0; i < 32; ++i) {
        base64_buffer[i] = 0;
    }
}

