#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <util/delay.h>

#include "uart0_no_rx.h"
#include "base64tx.h"

/*
 * This program generates some Base64 data.
 * It's the same data as used for testing keymap logic, but this program is intended
 * to create a library for Base64 transmitting of variable-bit-length data.
 * It reads the voltages on PA0, 2, 3, 4 and digital values on PA5, 6, 7, and PB0.
 * It sends the data every 100ms on PA1 in 460800-8-N-1 in Base64 format, without =.
 * The bit size of the data pieces is:
 * 4x 10 bit (ADC channels),
 * 4x 1 bit (Buttons)
 */

// State of ADC ISR:
uint8_t ADC_step;

// Let ADC read PA0, PA2, PA3, PA4, and push values to Base64 buffer:
ISR(ADC_vect)
{
    if (ADC_step == 1) {
        // Send PA0:
        push_value(ADC, 10);
        // Read PA2:
        ADC_step = 2;
        ADMUXA = 2;
        ADCSRA |= (1 << ADSC);
    } else if (ADC_step == 2) {
        // Send PA2;
        push_value(ADC, 10);
        // Read PA3:
        ADC_step = 3;
        ADMUXA = 3;
        ADCSRA |= (1 << ADSC);
    } else if (ADC_step == 3) {
        // Send PA3:
        push_value(ADC, 10);
        // Read PA4:
        ADC_step = 4;
        ADMUXA = 4;
        ADCSRA |= (1 << ADSC);
    } else {
        // Send PA4:
        push_value(ADC, 10);
        // Send digital values:
        uint8_t buttons = (((~PINA) & 0b11100000) >> 5)
            | (((~PINB) & 0b00000100) << 1);
        push_value(buttons, 4);
        send_buffer_base64();
        reset_buffer();
    }
}

// Timer1 times out every 100ms, then reading of ADC0 (PA0) is started:
ISR(TIMER1_COMPA_vect)
{
    // Read PA0:
    ADC_step = 1;
    ADMUXA = 0;
    ADCSRA |= (1 << ADSC);
}

int main(void)
{
    _delay_ms(3);

    // Clock source: 14.7456MHz crystal, no prescaling, no clock output.
    CCP = 0xd8;
    CLKCR = (1 << CSTR) | (1 << CKOUTC) | (0b1110 << CKSEL0);
    CCP = 0xd8;
    CLKPR = 0;

    // Timer1:
    // Period 100ms (with prescaler 1024):
    OCR1A = F_CPU / 10 / 1024;
    // OC1A Interrupt:
    TIMSK1 = (1 << OCIE1A);
    // CTC Mode with OC1A
    TCCR1B = (1 << WGM12) | (5 << CS10);

    // UART:
    uart_init(UART_BAUD_SELECT_DOUBLE_SPEED(460800, F_CPU));

    // ADC:
    // Enable, Interrupt, Prescaler 64:
    ADCSRA = (1 << ADEN) | (1 << ADIE) | (6 << ADPS0);

    // Pullup on digital inputs:
    PUEA = 0b11100000;
    PUEB = 0b00000100;

    // Execute:
    sei();
    while (1) {
    }
}

